import logging
from datetime import datetime, timedelta
from dateutil.parser import parse

import requests
from opentsdb_python_metrics.metric_wrappers import SendMetricMixin

from ocs_ingester.utils import metrics
from ocs_ingester.exceptions import BackoffRetryError, DoNotRetryError
from ocs_ingester.settings import settings as ingester_settings

from ocs_archive.settings import settings as archive_settings

logger = logging.getLogger('ocs_ingester')


def obs_end_time_from_dict(archive_record):
    obs_date = parse(archive_record.get('observation_date'))
    end_time = archive_record.get('headers', {}).get(archive_settings.OBSERVATION_END_TIME_KEY)
    if end_time:
        end_time = parse(end_time)
        # observation end is just a time - we need the date as well to be sure when this is
        end_date = obs_date.date()
        if abs(obs_date.hour - end_time.hour) > 12:
            # There was a date rollover during this observation, so set the date for utstop
            end_date += timedelta(days=1)

        return datetime.combine(end_date, end_time.time())
    elif archive_record.get('exposure_time'):
        return obs_date + timedelta(seconds=archive_record['exposure_time'])
    return obs_date


class ArchiveService(SendMetricMixin):
    def __init__(self, api_root, auth_token):
        self.api_root = api_root
        self.headers = {'Authorization': 'Token {}'.format(auth_token)}

    def handle_response(self, response):
        try:
            response.raise_for_status()
        except requests.exceptions.ConnectionError as exc:
            raise BackoffRetryError(exc)
        except requests.exceptions.HTTPError as exc:
            # HTTP 4xx errors are "client errors", meaning that the client sent
            # an incorrectly formatted request. There is no reason to retry an
            # incorrectly formatted request; it will only fail again.
            if 400 <= response.status_code < 500:
                raise DoNotRetryError(exc)

            # All other responses should back off and retry at a later time.
            # The most likely cause is that the HTTP server is unavailable
            # or overloaded, so we should try again later.
            raise BackoffRetryError(exc)

        # Return JSON data to client
        return response.json()

    def version_exists(self, md5):
        response = requests.get(
            '{0}versions/?md5={1}'.format(self.api_root, md5), headers=self.headers
        )
        result = self.handle_response(response)
        try:
            return result['count'] > 0
        except KeyError as e:
            raise BackoffRetryError(e)

    @metrics.method_timer('ingester.post_frame')
    def post_frame(self, archive_record):
        response = requests.post(
            '{0}frames/'.format(self.api_root), json=archive_record, headers=self.headers
        )
        result = self.handle_response(response)
        logger.info('Ingester posted frame to archive', extra={
            'tags': {
                'filename': result.get('filename'),
                'request_id': archive_record.get('request_id'),
                'proposal_id': result.get('proposal_id'),
                'id': result.get('id')
            }
        })
        # Add some useful information from the result
        archive_record['frameid'] = result.get('id')
        archive_record['filename'] = result.get('filename')
        archive_record['url'] = result.get('url')
        # Record metric for the ingest lag (time between date of image vs date ingested)
        ingest_lag = datetime.utcnow() - obs_end_time_from_dict(archive_record)
        self.send_metric(
            metric_name='ingester.ingest_lag',
            value=ingest_lag.total_seconds(),
            asynchronous=ingester_settings.SUBMIT_METRICS_ASYNCHRONOUSLY,
            **ingester_settings.EXTRA_METRICS_TAGS
        )
        return archive_record



import logging
from lcogt_logging import LCOGTFormatter


# Logging configuration, can import and setup to use.
logConf = {
    "formatters": {
        "default": {
            "()": LCOGTFormatter
        }
    },
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "formatter": "default"
        }
    },
    "loggers": {
        "ocs_ingester": {
            "handlers": ["console"],
            "level": logging.INFO,
            "propagate": False,
        }
    },
    "version": 1
}



#!/bin/env python3
"""
Command-line entrypoint to the ingester library.

Examples:

    See available options::

        (venv) ocs_ingest_frame --help

"""
import sys
import argparse

from ocs_ingester.ingester import frame_exists, upload_file_and_ingest_to_archive
from ocs_ingester.settings import settings
from ocs_ingester.exceptions import NonFatalDoNotRetryError

description = (
    'Upload a FITS file to the science archive of an observatory control system. This script will output the resulting '
    'URL if the upload is successful. An optional flag --check-only can be used to check for the existence of a file '
    'without uploading it (based on md5).'
)


def main():
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('path', help='Path to file')
    parser.add_argument('--api-root', help='API root')
    parser.add_argument('--auth-token', help='API token')
    parser.add_argument('--bucket', help='S3 bucket name')
    parser.add_argument('--process-name', help='Tag set in collected metrics')
    parser.add_argument('--check-only', action='store_true', help='Only check if the frame exists in the archive. \
                                                                   returns a status code of 0 if found, 1 if not \
                                                                   (or an error occurred)')
    args = parser.parse_args()

    # Submit metrics synchronously so that they all get submitted before the program exits
    settings.SUBMIT_METRICS_ASYNCHRONOUSLY = False

    if args.process_name:
        settings.EXTRA_METRICS_TAGS['ingester_process_name'] = args.process_name

    try:
        with open(args.path, 'rb') as fileobj:
            if args.check_only:
                try:
                    check_args = {
                        k: v for k, v in vars(args).items() if k in ['api_root', 'auth_token'] and v is not None
                    }
                    exists = frame_exists(fileobj, **check_args)

                except Exception as e:
                    sys.stdout.write(str(e))
                    sys.exit(1)
                sys.stdout.write(str(exists))
                sys.exit(int(not exists))

            try:
                ingest_args = {
                    k: v for k, v in vars(args).items() if k in ['api_root', 'auth_token', 'bucket'] and v is not None
                }
                result = upload_file_and_ingest_to_archive(fileobj=fileobj, path=args.path, **ingest_args)
            except NonFatalDoNotRetryError as e:
                sys.stdout.write(str(e))
                sys.exit(0)
            except Exception as e:
                sys.stdout.write('Exception uploading file: ')
                sys.stdout.write(str(e))
                sys.exit(1)

            sys.stdout.write(result['url'])
            sys.exit(0)

    except Exception as e:
        sys.stdout.write(str(e))
        sys.exit(1)


if __name__ == '__main__':
    main()



import functools
from datetime import datetime

from opentsdb_python_metrics.metric_wrappers import metric_timer_with_tags, send_tsdb_metric

from ocs_ingester.settings import settings as ingester_settings


def method_timer(metric_name):
    """Decorator to add extra tags to collected runtime metrics"""
    def method_timer_decorator(method):
        def wrapper(self, *args, **kwargs):
            # Decorate the wrapped method with metric_timer_with_tags, which does the work of figuring out
            # how long a method takes to run, so that the settings used are evaluated at runtime. An example
            # of when settings are changed at runtime is when the ingester command line entrypoint is used.
            @metric_timer_with_tags(
                metric_name=metric_name,
                asynchronous=ingester_settings.SUBMIT_METRICS_ASYNCHRONOUSLY,
                **ingester_settings.EXTRA_METRICS_TAGS
            )
            @functools.wraps(method)
            def run_method(self, *args, **kwargs):
                return method(self, *args, **kwargs)
            return run_method(self, *args, **kwargs)
        return wrapper
    return method_timer_decorator

@method_timer('ingester.upload_file')
def upload_and_collect_metrics(filestore_class, datafile):
    start_time = datetime.utcnow()
    # Upload the file to s3 and get version information back
    version = filestore_class.store_file(data_file=datafile)
    # Record metric for the bytes transferred / time to upload
    upload_time = datetime.utcnow() - start_time
    bytes_per_second = len(datafile.open_file) / upload_time.total_seconds()
    send_tsdb_metric(
        metric_name='ingester.s3_upload_bytes_per_second',
        value=bytes_per_second,
        asynchronous=ingester_settings.SUBMIT_METRICS_ASYNCHRONOUSLY,
        **ingester_settings.EXTRA_METRICS_TAGS
    )
    return version

@method_timer('ingester.get_md5')
def get_md5_and_collect_metrics(file_class):
    return file_class.get_md5()



"""
Exceptions raised by the ingester library.
"""


class DoNotRetryError(Exception):
    """
    Exception that is raised when an error happens that
    will undoubtedly repeat if called again. The task should
    not be retried.
    """
    pass


class RetryError(Exception):
    """
    Exception that is raised when an error happens that
    can be retried.
    """
    pass


class BackoffRetryError(Exception):
    """
    Exception that is raised when an error happens that
    can be retried with an expontential backoff. For example,
    networking latency errors that may succeed at a later time.
    """
    pass


class NonFatalDoNotRetryError(Exception):
    """
    Exception that is raised when an error happens that
    should not be retried and is also not a fatal condition.
    """
    pass



"""``ingester.py`` - Top level functions for adding data to the science archive.

Data is added to the science archive using the archive API and an S3 client. The steps necessary to add
data to the science archive are as follows:

    1) Check that the file does not yet exist in the science archive
    2) Validate and build a cleaned dictionary of the headers of the FITS file, or from provided metadata when uploading
       a non-FITS file
    3) Upload the file to S3
    4) Combine the results from steps 2 and 3 into a record to be added to the science archive database

Examples:
    Ingest a file one step at a time:

    >>> from ocs_ingester import ingester
    >>> with open('tst1mXXX-ab12-20191013-0001-e00.fits.fz', 'rb') as fileobj:
    >>>     if not ingester.frame_exists(fileobj):
    >>>        record = ingester.validate_fits_and_create_archive_record(fileobj)
    >>>        s3_version = ingester.upload_file_to_s3(fileobj)
    >>>        ingested_record = ingester.ingest_archive_record(s3_version, record)

    Ingest a file in one step:

    >>> from ocs_ingester import ingester
    >>> with open('tst1mXXX-ab12-20191013-0001-e00.fits.fz', 'rb') as fileobj:
    >>>    ingested_record = ingester.upload_file_and_ingest_to_archive(fileobj)

"""

from ocs_ingester.exceptions import BackoffRetryError, NonFatalDoNotRetryError, DoNotRetryError
from ocs_ingester.archive import ArchiveService
from ocs_ingester.utils.metrics import upload_and_collect_metrics, get_md5_and_collect_metrics
from ocs_ingester.settings import settings as ingester_settings

from ocs_archive.settings import settings as archive_settings
from ocs_archive.input.file import File, FileSpecificationException
from ocs_archive.input.filefactory import FileFactory
from ocs_archive.storage.filestorefactory import FileStoreFactory
from ocs_archive.storage.filestore import FileStoreSpecificationError, FileStoreConnectionError


def frame_exists(fileobj, api_root=ingester_settings.API_ROOT, auth_token=ingester_settings.AUTH_TOKEN):
    """Checks if the file exists in the science archive.

    Computes the md5 of the given file and checks whether a file with that md5 already exists in
    the science archive.

    Args:
        fileobj (file-like object): File-like object
        api_root (str): Science archive API root url
        auth_token (str): Science archive API authentication token

    Returns:
        bool: Boolean indicating whether the file exists in the science archive

    Raises:
        ocs_ingester.exceptions.BackoffRetryError: If there was a problem getting
            a response from the science archive API

    """
    archive = ArchiveService(api_root=api_root, auth_token=auth_token)
    md5 = get_md5_and_collect_metrics(File(fileobj))
    return archive.version_exists(md5)


def validate_fits_and_create_archive_record(fileobj, path=None, file_metadata=None,
                                            required_headers=archive_settings.REQUIRED_HEADERS,
                                            blacklist_headers=archive_settings.HEADER_BLACKLIST):
    """Validates the FITS file and creates a science archive record from it.

    Checks that required headers are present, removes blacklisted headers, and cleans other
    headers such that they are valid for ingestion into the science archive.

    Args:
        fileobj (file-like object): File-like object
        path (str): File path/name for this object. This option may be used to override the filename
            associated with the fileobj. It must be used if the fileobj does not have a filename.
        file_metadata (dict): Dictionary of file metadata to use when generating the archive record for a non-FITS file.
            This must be used when uploading a non-FITS file.
        required_headers (tuple): FITS headers that must be present
        blacklist_headers (tuple): FITS headers that should not be ingested

    Returns:
        dict: Constructed science archive record. For example::

            {
                'basename': 'tst1mXXX-ab12-20191013-0001-e00',
                'headers': {
                    'FILTER': 'rp',
                    'DATE-OBS': '2019-10-13T10:13:00',
                    ...
                },
                'area': ...
                ...
            }

    Raises:
        ocs_ingester.exceptions.DoNotRetryError: If required headers could not be found

    """
    try:
        open_file = File(fileobj, path)
        datafile = FileFactory.get_datafile_class_for_extension(open_file.extension)(
            open_file, file_metadata, blacklist_headers, required_headers
        )
    except FileSpecificationException as fe:
        raise DoNotRetryError(str(fe))
    json_record = datafile.get_header_data().get_archive_frame_data()
    json_record['headers'] = datafile.get_header_data().get_headers()
    json_record['area'] = datafile.get_wcs_corners()
    json_record['basename'] = open_file.basename
    return json_record


def upload_file_to_file_store(fileobj, path=None, file_metadata=None):
    """Uploads a file to the S3 bucket.

    Args:
        fileobj (file-like object): File-like object
        path (str): File path/name for this object. This option may be used to override the filename
            associated with the fileobj. It must be used if the fileobj does not have a filename.
        file_metadata (dict): Dictionary of file metadata to use when generating the archive record for a non-FITS file.
            This must be used when uploading a non-FITS file.

    Returns:
        dict: Version information for the file that was uploaded. For example::

            {
                'key': '792FE6EFFE6FAD7E',
                'md5': 'ECD9B357D67117BE8BF38D6F4B4A6',
                'extension': '.fits.fz'
            }

    Hint:
        The response contains an "md5" field, which is the md5 computed by file store. It is a good idea to
        check that this md5 is the same as the locally computed md5 of the file to make sure that
        the entire file was successfully uploaded.

    Raises:
        ocs_ingester.exceptions.BackoffRetryError: If there is a problem connecting to file store
        ocs_ingester.exceptions.DoNotRetryError: If there is a problem configuring file store
    """
    try:
        open_file = File(fileobj, path)
        datafile = FileFactory.get_datafile_class_for_extension(open_file.extension)(
            open_file, file_metadata
        )
    except FileSpecificationException as fe:
        raise DoNotRetryError(str(fe))

    try:
        filestore = FileStoreFactory.get_file_store_class()()
        # Returns the version, which holds in it the md5 that was uploaded
        return upload_and_collect_metrics(filestore, datafile)
    except FileStoreSpecificationError as fe:
        raise DoNotRetryError(str(fe))
    except FileStoreConnectionError as fce:
        raise BackoffRetryError(str(fce))


def ingest_archive_record(version, record, api_root=ingester_settings.API_ROOT,
                          auth_token=ingester_settings.AUTH_TOKEN):
    """Adds a record to the science archive database.

    Args:
        version (dict): Version information returned from the upload to S3
        record (dict): Science archive record to ingest
        api_root (str): Science archive API root url
        auth_token (str): Science archive API authentication token

    Returns:
        dict: The science archive record that was ingested. For example::

        {
            'basename': 'tst1mXXX-ab12-20191013-0001-e00',
            'version_set': [
                {
                    'key': '792FE6EFFE6FAD7E',
                    'md5': 'ECD9B357D67117BE8BF38D6F4B4A6',
                    'extension': '.fits.fz'
                    }
                ],
            'headers': {
                'REQNUM': 12345,
                ...
            }
            ...
        }

    Raises:
        ocs_ingester.exceptions.BackoffRetryError: If there was a problem connecting to the science archive
        ocs_ingester.exceptions.DoNotRetryError: If there was a problem with the record that must be fixed before
            attempting to ingest it again

    """
    archive = ArchiveService(api_root=api_root, auth_token=auth_token)
    # Construct final archive payload and post to archive
    record['version_set'] = [version]
    return archive.post_frame(record)


def upload_file_and_ingest_to_archive(fileobj, path=None, file_metadata=None,
                                      required_headers=archive_settings.REQUIRED_HEADERS,
                                      blacklist_headers=archive_settings.HEADER_BLACKLIST,
                                      api_root=ingester_settings.API_ROOT, auth_token=ingester_settings.AUTH_TOKEN):
    """Uploads a file to S3 and adds the associated record to the science archive database.

    This is a standalone function that runs all of the necessary steps to add data to the
    science archive.

    Args:
        fileobj (file-like object): File-like object
        path (str): File path/name for this object. This option may be used to override the filename
            associated with the fileobj. It must be used if the fileobj does not have a filename.
        file_metadata (dict): Dictionary of file metadata to use when generating the archive record for a non-FITS file.
            This must be used when uploading a non-FITS file.
        api_root (str): Science archive API root url
        auth_token (str): Science archive API authentication token
        required_headers (tuple): FITS headers that must be present
        blacklist_headers (tuple): FITS headers that should not be ingested

    Returns:
        dict: Information about the uploaded file and record. For example:

            {
                'basename': 'tst1mXXX-ab12-20191013-0001-e00',
                'version_set': [
                    {
                        'key': '792FE6EFFE6FAD7E',
                        'md5': 'ECD9B357D67117BE8BF38D6F4B4A6',
                        'extension': '.fits.fz'
                        }
                    ],
                'headers': {
                    'REQNUM': 12345,
                    ...
                }
                ...
            }

    Raises:
        ocs_ingester.exceptions.NonFatalDoNotRetryError: If the file already exists in the science archive
        ocs_ingester.exceptions.BackoffRetryError: If the md5 computed locally does not match the md5
            computed by S3, if there was an error connecting to S3, or if there was a problem reaching
            the science archive.
        ocs_ingester.exceptions.DoNotRetryError: If there was a problem that must be fixed before attempting
             to ingest again

    """
    try:
        if file_metadata is None:
            file_metadata = {}
        open_file = File(fileobj, path)
        datafile = FileFactory.get_datafile_class_for_extension(open_file.extension)(
            open_file, file_metadata, required_headers=required_headers, blacklist_headers=blacklist_headers
        )
        filestore = FileStoreFactory.get_file_store_class()()
    except (FileSpecificationException, FileStoreSpecificationError) as fe:
        raise DoNotRetryError(str(fe))

    archive = ArchiveService(api_root=api_root, auth_token=auth_token)
    ingester = Ingester(datafile, filestore, archive)
    return ingester.ingest()


class Ingester(object):
    """Ingest a single file into the archive.

    A single instance of this class is responsible for parsing a fits file,
    uploading the data to s3, and making a call to the archive api.
    """
    def __init__(self, datafile, filestore, archive):
        self.datafile = datafile
        self.filestore = filestore
        self.archive = archive

    def ingest(self):
        # Get the Md5 checksum of this file and check if it already exists in the archive
        md5 = get_md5_and_collect_metrics(self.datafile.open_file)
        if self.archive.version_exists(md5):
            raise NonFatalDoNotRetryError('Version with this md5 already exists')

        # Upload the file to s3 and get version information back
        version = upload_and_collect_metrics(self.filestore, self.datafile)

        # Make sure our md5 matches amazons
        if version['md5'] != md5:
            raise BackoffRetryError('S3 md5 did not match ours')

        # Construct final archive payload and post to archive
        record = self.datafile.get_header_data().get_archive_frame_data()
        record['headers'] = self.datafile.get_header_data().get_headers()
        record['area'] = self.datafile.get_wcs_corners()
        record['version_set'] = [version]
        record['basename'] = self.datafile.open_file.basename
        return self.archive.post_frame(record)



from setuptools import setup

# Read the contents of the README
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='ocs-ingester',
    version='3.0.4',
    description='Ingest frames into the science archive of an observatory control system',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/observatorycontrolsystem/ingester',
    packages=['ocs_ingester', 'ocs_ingester.utils', 'ocs_ingester.settings', 'ocs_ingester.scripts'],
    python_requires='>=3.6',
    classifiers=[
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)'
    ],
    install_requires=[
        'astropy',
        'requests',
        'boto3',
        'ocs_archive==0.2.9',
        'python-dateutil',
        'lcogt-logging',
        'opentsdb-python-metrics>=0.2.0'
    ],
    extras_require={
        'tests': ['pytest']
    },
    entry_points={
        'console_scripts': [
            'ocs_ingest_frame = ocs_ingester.scripts.ingest_frame:main',
        ]
    }
)



import os
import ast


# General settings
API_ROOT = os.getenv('API_ROOT', 'http://127.0.0.1:8000/')
AUTH_TOKEN = os.getenv('AUTH_TOKEN', 'c158b4f055c5abdd9f520c8501159478f6f738ac')

# Whether to submit the metrics asynchronously
SUBMIT_METRICS_ASYNCHRONOUSLY = ast.literal_eval(os.getenv('SUBMIT_METRICS_ASYNCHRONOUSLY', 'False'))

# Extra tags for metrics
EXTRA_METRICS_TAGS = {
    'ingester_process_name': os.getenv('INGESTER_PROCESS_NAME', 'ingester')
}
